include dll.e
include machine.e
without warning
without type_check

global constant
SBYTE   = #01000001,
BYTE    = #02000001,
SWORD   = #01000002, 
WORD    = #02000002, 
SDWORD  = #01000004,
DWORD   = #02000004,
FLOAT   = #03000004,
REAL4   = #03000004,
REAL8   = #03000008

global function define_asm_func(object code, sequence proto, integer rettyp, integer align)
atom addr
    if atom(code) then
        addr=code
    else
        addr=allocate(length(code)+align)
        if align>0 then
            if remainder(addr,align) then
                addr += align-remainder(addr,align)
            end if
        end if
    end if
    poke(addr,code)
    return define_c_func("", addr, proto, rettyp)
end function

global function define_asm_proc(object code, sequence proto, integer align)
atom addr
    if atom(code) then
        addr=code
    else
        addr=allocate(length(code)+align)
        if align>0 then
            if remainder(addr,align) then
                addr += align-remainder(addr,align)
            end if
        end if
    end if
    poke(addr,code)
    return define_c_proc("", addr, proto)
end function

-- These are the standard prolog and epilog for a windows procedure

global function ENTER_PROC(integer nbytes) -- number of bytes to link for locals
sequence s
    s = {
    #55,                        -- push   ebp        ; create stack frame
    #8B,#EC                     -- mov    ebp,esp    ; same effect as enter 0,0
    }
    if nbytes then
        s &= 
        #ED                     -- sub    ebp,imm32  ; link local storage
        & int_to_bytes(nbytes)
    end if
    -- windows requires the following registers to be saved
    -- pusha is convenient but popa destroys the eax
    s &= {      
    #52,                        -- push   edx
    #51,                        -- push   ecx
    #53,                        -- push   ebx
    #56,                        -- push   esi
    #57                         -- push   edi
    }
    return s
end function

-- Strictly speaking you are only required to save these registers for a
-- callback but 90% of the time you need most if not all of them anyways.
-- Virtually all programming languages would require this also.
-- For euphoria you should push any register except eax that you intend to use. 

global function LEAVE_PROC(integer nbytes) -- number of bytes (parameters on stack)
sequence s
    s= {
    #5F,                        -- pop    edi
    #5E,                        -- pop    esi
    #5B,                        -- pop    ebx
    #59,                        -- pop    ecx
    #5A,                        -- pop    edx
    #C9,                        -- leave            ; destroy stack frame
    #C2                         -- ret    imm16
    } & and_bits(nbytes,#FF) & floor(nbytes/#100)
    return s
end function

global function RETURN(integer nbytes)
sequence s
    if nbytes then
        nbytes=and_bits(nbytes,#FFFF)
        s= {
        #C2                         -- ret imm16
        } & and_bits(nbytes,#FF) & floor(nbytes/#100)
    else
        s={#C3}
    end if
    return s
end function


--------------------------------------------------------------------------
--------------------------------------------------------------------------
-- Bit and Byte functions
--------------------------------------------------------------------------
--------------------------------------------------------------------------


sequence code_string


--------------------------------------------------------------------------
-- Rotate
--------------------------------------------------------------------------

code_string =           -- number:DWORD positions:DWORD
{
    #51,                    -- push  ecx
    #8B,#44,#24,#08,        -- mov   eax,[esp+8]
    #8B,#4C,#24,#0C,        -- mov   ecx,[esp+12]
    #D3,#C0,                -- rol   eax,cl
    #59,                    -- pop   ecx
    #C2,#08,#00             -- ret   8
}
global constant ROL_D = define_asm_func(code_string,{DWORD,DWORD},DWORD,4)

global function rol_d(atom n, integer p)
    return c_func(ROL_D,{n,p})
end function

--------------------------------------------------------------------------

code_string =           -- number:DWORD positions:DWORD
{
    #51,                    -- push  ecx
    #8B,#44,#24,#08,        -- mov   eax,[esp+8]
    #8B,#4C,#24,#0C,        -- mov   ecx,[esp+12]
    #D3,#C8,                -- ror   eax,cl
    #59,                    -- pop   ecx
    #C2,#08,#00             -- ret   8
}
global constant ROR_D = define_asm_func(code_string,{DWORD,DWORD},DWORD,4)

global function ror_d(atom n, integer p)
    return c_func(ROR_D,{n,p})
end function

--------------------------------------------------------------------------
code_string =           -- number:DWORD positions:DWORD
{
    #51,                    -- push  ecx
    #8B,#44,#24,#08,        -- mov   eax,[esp+8]
    #8B,#4C,#24,#0C,        -- mov   ecx,[esp+12]
    #66,#D3,#C0,            -- rol   ax,cl
    #59,                    -- pop   ecx
    #C2,#08,#00             -- ret   8
}
global constant ROL_W = define_asm_func(code_string,{DWORD,DWORD},DWORD,4)

global function rol_w(atom n, integer p)
    return c_func(ROL_W,{n,p})
end function

--------------------------------------------------------------------------

code_string =           -- number:DWORD positions:DWORD
{
    #51,                    -- push  ecx
    #8B,#44,#24,#08,        -- mov   eax,[esp+8]
    #8B,#4C,#24,#0C,        -- mov   ecx,[esp+12]
    #66,#D3,#C8,            -- ror   ax,cl
    #59,                    -- pop   ecx
    #C2,#08,#00             -- ret   8
}
global constant ROR_W = define_asm_func(code_string,{DWORD,DWORD},DWORD,4)

global function ror_w(atom n, integer p)
    return c_func(ROR_W,{n,p})
end function

--------------------------------------------------------------------------
code_string =           -- number:DWORD positions:DWORD
{
    #51,                    -- push  ecx
    #8B,#44,#24,#08,        -- mov   eax,[esp+8]
    #8B,#4C,#24,#0C,        -- mov   ecx,[esp+12]
    #D2,#C0,                -- rol   al,cl
    #59,                    -- pop   ecx
    #C2,#08,#00             -- ret   8
}
global constant ROL_B = define_asm_func(code_string,{DWORD,DWORD},DWORD,4)

global function rol_b(atom n, integer p)
    return c_func(ROL_B,{n,p})
end function

--------------------------------------------------------------------------

code_string =           -- number:DWORD positions:DWORD
{
    #51,                    -- push  ecx
    #8B,#44,#24,#08,        -- mov   eax,[esp+8]
    #8B,#4C,#24,#0C,        -- mov   ecx,[esp+12]
    #D2,#C8,                -- ror   al,cl
    #59,                    -- pop   ecx
    #C2,#08,#00             -- ret   8
}
global constant ROR_B = define_asm_func(code_string,{DWORD,DWORD},DWORD,4)

global function ror_b(atom n, integer p)
    return c_func(ROR_B,{n,p})
end function

--------------------------------------------------------------------------
-- logical Shift
--------------------------------------------------------------------------

global sequence pwrs2 pwrs2={}

for i=0 to 32 do 
    pwrs2=append(pwrs2,power(2,i))
end for

--------------------------------------------------------------------------

code_string =           -- number:DWORD positions:DWORD
{
    #51,                    -- push  ecx
    #8B,#44,#24,#08,        -- mov   eax,[esp+8]
    #8B,#4C,#24,#0C,        -- mov   ecx,[esp+12]
    #D3,#E0,                -- shl   eax,cl
    #59,                    -- pop   ecx
    #C2,#08,#00             -- ret   8
}
global constant SHL_D = define_asm_func(code_string,{DWORD,DWORD},DWORD,4)

global function shl_d(atom n, integer p)
    return c_func(SHL_D,{n,p})
end function

--------------------------------------------------------------------------

code_string =           -- number:DWORD positions:DWORD
{
    #51,                    -- push  ecx
    #8B,#44,#24,#08,        -- mov   eax,[esp+8]
    #8B,#4C,#24,#0C,        -- mov   ecx,[esp+12]
    #D3,#E8,                -- shr   eax,cl
    #59,                    -- pop   ecx
    #C2,#08,#00             -- ret   8
}
global constant SHR_D = define_asm_func(code_string,{DWORD,DWORD},DWORD,4)

global function shr_d(atom n, integer p)
    return c_func(SHR_D,{n,p})
end function

--------------------------------------------------------------------------

code_string =           -- number:DWORD positions:DWORD
{
    #51,                    -- push  ecx
    #8B,#44,#24,#08,        -- mov   eax,[esp+8]
    #8B,#4C,#24,#0C,        -- mov   ecx,[esp+12]
    #66,#D3,#E0,            -- shl   ax,cl
    #59,                    -- pop   ecx
    #C2,#08,#00             -- ret   8
}
global constant SHL_W = define_asm_func(code_string,{DWORD,DWORD},DWORD,4)

global function shl_w(atom n, integer p)
    return c_func(SHL_W,{n,p})
end function

--------------------------------------------------------------------------

code_string =           -- number:DWORD positions:DWORD
{
    #51,                    -- push  ecx
    #8B,#44,#24,#08,        -- mov   eax,[esp+8]
    #8B,#4C,#24,#0C,        -- mov   ecx,[esp+12]
    #66,#D3,#E8,            -- shr   ax,cl
    #59,                    -- pop   ecx
    #C2,#08,#00             -- ret   8
}
global constant SHR_W = define_asm_func(code_string,{DWORD,DWORD},DWORD,4)

global function shr_w(atom n, integer p)
    return c_func(SHR_W,{n,p})
end function

--------------------------------------------------------------------------

code_string =           -- number:DWORD positions:DWORD
{
    #51,                    -- push  ecx
    #8B,#44,#24,#08,        -- mov   eax,[esp+8]
    #8B,#4C,#24,#0C,        -- mov   ecx,[esp+12]
    #D2,#E0,                -- shl   al,cl
    #59,                    -- pop   ecx
    #C2,#08,#00             -- ret   8
}
global constant SHL_B = define_asm_func(code_string,{DWORD,DWORD},DWORD,4)

global function shl_b(atom n, integer p)
    return c_func(SHL_B,{n,p})
end function

--------------------------------------------------------------------------

code_string =           -- number:DWORD positions:DWORD
{
    #51,                    -- push  ecx
    #8B,#44,#24,#08,        -- mov   eax,[esp+8]
    #8B,#4C,#24,#0C,        -- mov   ecx,[esp+12]
    #D2,#E8,                -- shr   al,cl
    #59,                    -- pop   ecx
    #C2,#08,#00             -- ret   8
}
global constant SHR_B = define_asm_func(code_string,{DWORD,DWORD},DWORD,4)

global function shr_b(atom n, integer p)
    return c_func(SHR_B,{n,p})
end function

--------------------------------------------------------------------------
-- Bit scan
--------------------------------------------------------------------------

code_string =           -- number:DWORD
{
    #B8,#FF,#FF,#FF,#FF,        -- mov   eax,-1
    #0F,#BC,#44,#24,#04,        -- bsr   eax,[esp+4]
    #C2,#04,#00                 -- ret   00004h
}
global constant BIT_SCANF = define_asm_func(code_string,{DWORD},SDWORD,4)

global function bit_scanf(atom n) -- find least significant set bit
    return c_func(BIT_SCANF,{n})
end function

--------------------------------------------------------------------------

code_string =           -- number:DWORD
{
    #B8,#FF,#FF,#FF,#FF,        -- mov   eax,-1
    #0F,#BD,#44,#24,#04,        -- bsr   eax,[esp+4]
    #C2,#04,#00                 -- ret   00004h
}
global constant BIT_SCANR = define_asm_func(code_string,{DWORD},SDWORD,4)

global function bit_scanr(atom n) -- find most significant set bit
    return c_func(BIT_SCANR,{n})
end function

--------------------------------------------------------------------------
-- Bit test & set
--------------------------------------------------------------------------

code_string =           -- number:DWORD, bit:DWORD
{
    #8B,#44,#24,#08,            -- mov   eax,[esp+8] ; bit
    #0F,#A3,#44,#24,#04,        -- bt    [esp+4],eax ; number
    #73,#08 ,                   -- jnc   @F
    #B8,#01,#00,#00,#00,        -- mov   eax,1
    #C2,#08,#00,                -- ret   8
                                -- @@:
    #B8,#00,#00,#00,#00,        -- mov   eax,0
    #C2,#08,#00                 -- ret   8
}
global constant BIT_TEST = define_asm_func(code_string,{DWORD,DWORD},DWORD,4)

global function bit_test(atom n, integer bit)
    return c_func(BIT_TEST,{n,bit})
end function

--------------------------------------------------------------------------

code_string =           -- number:DWORD bit:DWORD
{
   #53,                        -- push   ebx
   #8B,#44,#24,#08,            -- mov    eax,[esp+8]
   #8B,#5C,#24,#0C,            -- mov    ebx,[esp+12]
   #0F,#B3,#D8,                -- btr    eax,ebx
   #5B,                        -- pop    ebx
   #C2,#08,#00                 -- ret    8
}
global constant CLR_BIT = define_asm_func(code_string,{DWORD,DWORD},DWORD,4)

global function clr_bit(atom n, integer bit)
    return c_func(CLR_BIT,{n,bit})
end function

--------------------------------------------------------------------------

code_string =           --  number:DWORD bit:DWORD
{
   #53,                        -- push   ebx
   #8B,#44,#24,#08,            -- mov    eax,[esp+8]
   #8B,#5C,#24,#0C,            -- mov    ebx,[esp+12]
   #0F,#AB,#D8,                -- bts    eax,ebx
   #5B,                        -- pop    ebx
   #C2,#08,#00                 -- ret    8
}
global constant SET_BIT = define_asm_func(code_string,{DWORD,DWORD},DWORD,4)

global function set_bit(atom n, integer bit)
    return c_func(SET_BIT,{n,bit})
end function

--------------------------------------------------------------------------

code_string =           -- number:DWORD bit:DWORD
{
   #53,                        -- push   ebx
   #8B,#44,#24,#08,            -- mov    eax,[esp+8]
   #8B,#5C,#24,#0C,            -- mov    ebx,[esp+12]
   #0F,#BB,#D8,                -- btc    eax,ebx
   #5B,                        -- pop    ebx
   #C2,#08,#00                 -- ret    8
}
global constant NOT_BIT = define_asm_func(code_string,{DWORD,DWORD},DWORD,4)

global function not_bit(atom n, integer bit)
    return c_func(NOT_BIT,{n,bit})
end function

--------------------------------------------------------------------------
-- Byte swap
--------------------------------------------------------------------------

code_string =           -- number:DWORD
{
    #8B,#44,#24,#04,        -- mov   eax,[esp+4]
    #0F,#C8,                -- bswap eax
    #C2,#04,#00             -- ret   00004h
}
global constant BSWP_D = define_asm_func(code_string,{DWORD},DWORD,4)

global function bswp_d(atom n) -- bits 0-7 <-> bits 24-31  bits 8-15 <-> bits 16-23
    return c_func(BSWP_D,{n})
end function

--------------------------------------------------------------------------

code_string =           -- number:DWORD
{
    #8B,#44,#24,#04,        -- mov   eax,[esp+4]
    #86,#C4,                -- xchg  al,ah
    #C2,#04,#00             -- ret   00004h
}
global constant BSWP_W = define_asm_func(code_string,{DWORD},DWORD,4)

global function bswp_w(atom n) -- bits 0-7 <-> bits 8-15
    return c_func(BSWP_W,{n})
end function

--------------------------------------------------------------------------
-- Word swap
--------------------------------------------------------------------------

code_string =           -- number:DWORD
{
    #8B,#44,#24,#04,        -- mov   eax,[esp+4]
    #C1,#C0,#10,            -- ror   eax,16
    #C2,#04,#00             -- ret   00004h
}
global constant WSWP_D = define_asm_func(code_string,{DWORD},DWORD,4)

global function wswp_d(atom n) -- bits 0-15 <-> bits 16-31
    return c_func(WSWP_D,{n})
end function

--------------------------------------------------------------------------
-- CPUID
--------------------------------------------------------------------------

code_string =           -- input:DWORD, buffer:DWORD
{
   #55,                    -- push   ebp
   #8B,#EC,                -- mov    ebp, esp
   #60,                    -- pushad
					
   #B8,#00,#00,#00,#80,    -- mov eax,80000000h ; extended
   #39,#45,#08,            -- cmp input,eax ; if input < eax
   #73,#05,                -- jae @F
   #B8,#00,#00,#00,#00,    -- mov eax,0 ; standard
                           -- @@:
   #0F,#A2,                -- cpuid ; get max input
   #39,#45,#08,            -- cmp input,eax ; if input <= eax
   #77,#1B,                -- ja @fail

   #8B,#45,#08,            -- mov eax,input
   #0F,#A2,                -- cpuid
   #8B,#7D,#0C,            -- mov edi,buffer ; pointer to 16 byte buffer
   #FC,                    -- cld
   #AB,                    -- stosd    
   #8B,#C3,                -- mov eax,ebx
   #AB,                    -- stosd    
   #8B,#C1,                -- mov eax,ecx
   #AB,                    -- stosd    
   #8B,#C2,                -- mov eax,edx
   #AB,                    -- stosd
   #61,                    -- popad    
   #B8,#01,#00,#00,#00,    -- mov eax,TRUE
   #EB,#05,                -- jmp @F
                           -- @fail:
   #B8,#00,#00,#00,#00,    -- mov eax,FALSE
                           -- @@:
   #C9,                    -- leave  
   #C2,#08,#00             -- ret    00008h
}
global constant CPU_ID = define_asm_func(code_string,{DWORD,DWORD},DWORD,0)

global function cpu_id(atom input)
 atom buffer
 integer bool
 sequence regs
    buffer = allocate(16) poke4(buffer,{0,0,0,0}) regs = {}
    bool = c_func(CPU_ID,{input,buffer})
    regs = peek4u({buffer,4})
    free(buffer)
    return {bool,regs}
end function

-- input 0
-- EAX  maximum input value
-- EBX  Genu |                       Auth |
-- ECX  ntel |                       cAMD |
-- EDX  ineI |-- intel processors    enti |-- amd processors
-- Some strings that might be returned by cpu_id(0)
    -- "GenuineIntel" - Intel
    -- "AuthenticAMD" - AMD
    -- "CyrixInstead" - Cyrix
    -- "CentaurHauls" - Centaur
    -- "SiS SiS SiS " - SiS
    -- "NexGenDriven" - NexGen
    -- "GenuineTMx86", "TransmetaCPU" - Transmeta
    -- "RiseRiseRise" - Rise
    -- "UMC UMC UMC " - UMC
    -- "Geode by NSC" - National Semiconductor

-- input 1      pg. 151 Intel  Instruction Set Reference
-- EAX - version info
-- bits
-- 0-3     stepping id
-- 4-7     model
-- 8-11    family
-- 12-13   processor type

-- edx - feature info
-- bit
-- 0      FPU      contains an FPU
-- 1      VME      supports virtual-8086 mode enhancements
-- 2      DE       supports I/O breakpoints
-- 3      PSE      supports 4-Mbyte pages
-- 4      TSC      supports the RDTSC (read time stamp counter)
-- 5      MSR      supports the RDMSR (read model-specific register)
-- 6      PAE      supports physical addresses greater than 32 bits
-- 7      NCE      supports the CR4.MCE bit, enabling machine check exceptions
-- 8      CX8      supports the CMPXCHG8B (compare and exchange 8 bytes)
-- 9      APIC     contains an on-chip Advanced Programmable Interrupt Controller
-- 10     reserved
-- 11     SEP      supports the Fast System Calls, SYSENTER and SYSEXIT
-- 12     MTRR     supports machine-specific memory-type range registers
-- 13     PGEPTE  supports the CR4.PGE flag
-- 14     MCA      supports the MCG_CAP (machine check global capability)
-- 15     CMOV     supports the CMOV cc instruction, if (bit 0) FCMOV cc and FCOMI
-- 16     FGPAT    page attribute table, as above
-- 17     PSE-36   supports 4MB pages with 36 bit physical addresses.
-- 18     PN       supports the 96-bit Processor Number feature, feature is enabled
-- 19-22  Reserved
-- 23     MMX      supports the MMX instruction set
-- 24     FXSR     supports the FXSAVE and FXRSTOR instructions
-- 25     XMM      supports the Streaming SIMD Extensions instruction set
-- 26-31  Reserved

-- input 2 clones may not support input > 1 
-- (AL)   indicates the number of times CPUID must be executed with an input
--        value of 2 to get a complete description of the processor caches and TLBs.

-- (bit 31) of each register is cleared if register contains valid information

-- If a register contains valid information,
-- the information is contained in the low byte.

-- 00H  Null descriptor
-- 01H  Instruction TLB: 4K-Byte Pages, 4-way set associative, 32 entries
-- 02H  Instruction TLB: 4M-Byte Pages, fully associative, two entries
-- 03H  Data TLB: 4K-Byte Pages, 4-way set associative, 64 entries
-- 04H  Data TLB: 4M-Byte Pages, 4-way set associative, eight entries
-- 06H  Instruction cache: 8K Bytes, 4-way set associative, 32 byte line size
-- 08H  Instruction cache: 16K Bytes, 4-way set associative, 32 byte line size
-- 0AH  Data cache: 8K Bytes, 2-way set associative, 32 byte line size
-- 0CH  Data cache: 16K Bytes, 2-way or 4-way set associative, 32 byte line size
-- 40H  No L2 Cache
-- 41H  L2 Unified cache: 128K Bytes, 4-way set associative, 32 byte line size
-- 42H  L2 Unified cache: 256K Bytes, 4-way set associative, 32 byte line size
-- 43H  L2 Unified cache: 512K Bytes, 4-way set associative, 32 byte line size
-- 44H  L2 Unified cache: 1M Byte, 4-way set associative, 32 byte line size
-- 45H  L2 Unified cache: 2M Byte, 4-way set associative, 32 byte line size

-- Extended cpuid levels
-- input #80000000         maximum supported extended level and vendor ID string
-- input #80000001         processor family/model/stepping and features flags
-- input #80000002         processor name string (part 1)
-- input #80000003         processor name string (part 2)
-- input #80000004         processor name string (part 3)
-- input #80000005         L1 cache and TLB configuration 
-- input #80000006         L2 cache configuration
-- input #80000007         enhanced power management (EPM)
-- input #80000008         miscellaneous information
-- input #8000000A         SVM information
-- Extended levels greater than #8000000A are vendor specific.

-- See this page for more information on cpuid output.
-- http://www.sandpile.org/ia32/cpuid.htm

--function processor_brand_string()
--	sequence t,s
--	atom n
--	n=#80000002 s={}
--	for i=0 to 2 do
--		t=cpu_id(n+i)
--		for j=1 to 4 do
--			s&=int_to_bytes(t[2][j])
--		end for
--	end for
--	for i=1 to length(s) do
--		if s[i]=0 then
--			s=s[1..i-1] exit
--		end if	
--	end for
--	return s
--end function

--------------------------------------------------------------------------
-- Block Move
--------------------------------------------------------------------------

code_string =
                        -- align 16
ENTER_PROC(0) &
{                   
   #8B,#7D,#08,             -- mov    edi,destination
   #8B,#75,#0C,             -- mov    esi,source
   #8B,#4D,#10,             -- mov    ecx,ccnt

   #FC,                     -- cld                  ; set dir for move down
   #33,#DB,                 -- xor    ebx,ebx       ; adjustment for addr regs
                            -- .if esi < edi        ; if source < destination
   #3B,#F7,                 -- cmp    esi, edi
   #73,#0E,                 -- jae    @01
   #8D,#74,#31,#FC,         -- lea  esi,[esi+ecx-4]
   #8D,#7C,#39,#FC,         -- lea  edi,[edi+ecx-4]
   #BB,#03,#00,#00,#00,     --  mov ebx,3
   #FD,                     -- std                  ; set dir for move up
                            -- .endif
                            -- @01:

   #8B,#C1,                 -- mov    eax,ecx
   #C1,#E9,#06,             -- shr    ecx,6         ; divide by 64 = number of blocks
   #8B,#D1,                 -- mov    edx,ecx
   #C1,#E2,#06,             -- shl    edx,6         ; number of bytes to move in 64 byte blocks
   #2B,#C2,                 -- sub    eax,edx       ; number of bytes left

                            -- .if ecx != 0         ; move blocks of 64 bytes
   #0B,#C9,                 -- or     ecx, ecx
   #74,#1E,                 -- je     @02

   #90,#90,#90,#90,#90,     -- align 16 (nop)
   #90,#90,#90,#90,#90,
                            -- @@:
   #A5,                     -- movsd
   #A5,                     -- movsd
   #A5,                     -- movsd
   #A5,                     -- movsd
   #A5,                     -- movsd
   #A5,                     -- movsd
   #A5,                     -- movsd
   #A5,                     -- movsd
   #A5,                     -- movsd
   #A5,                     -- movsd
   #A5,                     -- movsd
   #A5,                     -- movsd
   #A5,                     -- movsd
   #A5,                     -- movsd
   #A5,                     -- movsd
   #A5,                     -- movsd
   #49,                     -- dec    ecx
   #75,#ED,                 -- jnz    @B
                            -- .endif
                            -- @02:

   #03,#F3,                 -- add    esi,ebx       ; adjust pointers
   #03,#FB,                 -- add    edi,ebx

   #8B,#C8,                 -- mov    ecx,eax       ; bytes to move
   #F3,#A4,                 -- rep    movsb
   #FC                      -- cld
} &
LEAVE_PROC(12)

global constant BLKMOV = define_asm_proc(code_string,{DWORD,DWORD,DWORD},16)

global procedure blkmov(atom destination, atom source, atom ccnt)
    c_proc(BLKMOV,{destination,source,ccnt})
end procedure

--------------------------------------------------------------------------
-- Debug
--------------------------------------------------------------------------

-- http://www.ollydbg.de/
-- this is a very good debugger and it's free
-- The best way to debug asm code in euphoria is to just put a #CC,
-- or int3, at the beginning of the routine to be debugged.
-- If you use ICEBP or BREAK you will be tracing for a while before
-- you get to your code.
-- As a rule ICE is used by hardware debuggers 'InCircuitEmulator',
-- however in the past I have used software debuggers that trapped
-- int1 but not int3, so I've included it.

global constant
int1 = #F1,
int3 = #CC

code_string = 
{  #F1,                -- INT1    ; ICEBP
   #C3                 -- RET
}
global constant ICEBP = define_asm_proc(code_string,{},0)

global procedure icebp()
    c_proc(ICEBP,{})
end procedure


code_string = 
{  #CC,                -- INT3    ; BREAK
   #C3                 -- RET
}
global constant BREAK = define_asm_proc(code_string,{},0)

global procedure break()
    c_proc(BREAK,{})
end procedure

--------------------------------------------------------------------------
-- byte word dword char string search
--------------------------------------------------------------------------

code_string =           -- b:DWORD, addr:DWORD, ccnt:DWORD
ENTER_PROC(0) &
{
   #8B,#45,#08,             -- mov    eax,b
   #8B,#4D,#10,             -- mov    ecx,ccnt
   #8B,#7D,#0C,             -- mov    edi,addr
   #F2,#AE,                 -- repne  scasb    ; search until match or ecx=0
   #B8,#00,#00,#00,#00,     -- mov    eax,0 
   #75,#05,                 -- jne    @F
   #8B,#C7,                 -- mov    eax,edi
   #83,#E8,#01              -- sub    eax,1
                            -- @@:
} &
LEAVE_PROC(12)
global constant BYTE_SCAN = define_asm_func(code_string,{DWORD,DWORD,DWORD},DWORD,4)

global function byte_scan(integer b, atom addr, integer ccnt)
    return c_func(BYTE_SCAN,{b,addr,ccnt})
end function

--------------------------------------------------------------------------

code_string =           -- w:DWORD, addr:DWORD, ccnt:DWORD
ENTER_PROC(0) &
{
   #8B,#45,#08,             -- mov    eax,w
   #8B,#4D,#10,             -- mov    ecx,ccnt
   #8B,#7D,#0C,             -- mov    edi,addr
   #C1,#E9,#01,             -- shr    ecx,1
   #F2,#66,#AF,             -- repne  scasw    ; search until match or ecx=0
   #B8,#00,#00,#00,#00,     -- mov    eax,0 
   #75,#05,                 -- jne    @F
   #8B,#C7,                 -- mov    eax,edi
   #83,#E8,#02              -- sub    eax,2
                            -- @@:
} &
LEAVE_PROC(12)
global constant WORD_SCAN = define_asm_func(code_string,{DWORD,DWORD,DWORD},DWORD,4)

global function word_scan(integer w, atom addr, integer ccnt)
    return c_func(WORD_SCAN,{w,addr,ccnt})
end function

--------------------------------------------------------------------------

code_string =           -- d:DWORD, addr:DWORD, ccnt:DWORD
ENTER_PROC(0) &
{
   #8B,#45,#08,             -- mov    eax,d
   #8B,#4D,#10,             -- mov    ecx,ccnt
   #8B,#7D,#0C,             -- mov    edi,addr
   #C1,#E9,#02,             -- shr    ecx,2
   #F2,#AF,                 -- repne  scasd    ; search until match or ecx=0
   #B8,#00,#00,#00,#00,     -- mov    eax,0 
   #75,#05,                 -- jne    @F
   #8B,#C7,                 -- mov    eax,edi
   #83,#E8,#04              -- sub    eax,4
                            -- @@:
} &
LEAVE_PROC(12)
global constant DWORD_SCAN = define_asm_func(code_string,{DWORD,DWORD,DWORD},DWORD,4)

global function dword_scan(atom d, atom addr, integer ccnt)
    return c_func(DWORD_SCAN,{d,addr,ccnt})
end function

--------------------------------------------------------------------------

-- this may look a bit strange but there is little difference
-- for case insensitive searches, and it's a lot faster for 
-- case sensitive searches or non alpha characters.

code_string =           -- c:DWORD, adr:DWORD, ccnt:DWORD, fcase:DWORD
ENTER_PROC(0) &
{
   #8B,#45,#08,             -- mov    eax,chr     ; char to find
   #8B,#7D,#0C,             -- mov    edi,adr     ; string to search
   #8B,#4D,#10,             -- mov    ecx,ccnt    ; loop counter
   #8B,#55,#14,             -- mov    edx,fcase   ; flag

   #F2,#AE,                 -- repne  scasb       ; search until match or ecx=0
   #74,#20,                 -- je     match

   #85,#D2,                 -- test   edx,edx     ; case flag=0 done
   #74,#21,                 -- jz    nomatch

   #3C,#41,                 -- cmp    al,'A'      ; if char is not alpha done
   #72,#1D,                 -- jb     nomatch
   #3C,#7A,                 -- cmp    al,'z'
   #77,#19,                 -- ja     nomatch
   #3C,#61,                 -- cmp    al,'a'
   #73,#04,                 -- jae    @F
   #3C,#5A,                 -- cmp    al,'Z'
   #77,#11,                 -- ja     nomatch
                            -- @@:                 ; if char is alpha and case<>0
   #34,#20,                 -- xor     al,20h      ; toggle case
   #8B,#7D,#0C,             -- mov    edi,adr
   #8B,#4D,#10,             -- mov    ecx,ccnt
   #F2,#AE,                 -- repne  scasb        ; search again
   #75,#05,                 -- jne    nomatch

                            -- match:
   #8B,#C7,                 -- mov    eax,edi
   #48,                     -- dec    eax
   #EB,#05,                 -- jmp    @F

                            -- nomatch:
   #B8,#00,#00,#00,#00      -- mov    eax,0
                            -- @@:
} &
LEAVE_PROC(16)
global constant CHAR_SCAN = define_asm_func(code_string,{DWORD,DWORD,DWORD,DWORD},DWORD,4)

global function char_scan(integer c, atom addr, atom ccnt, integer fcase)
    return c_func(CHAR_SCAN,{c,addr,ccnt,fcase})
end function

--------------------------------------------------------------------------

code_string =    -- a1:DWORD, a2:DWORD, count:DWORD, fcase:DWORD
ENTER_PROC(0) &
{
   #8B,#75,#08,         -- mov    esi,a1           ; string 1
   #8B,#7D,#0C,         -- mov    edi,a2           ; string 2
   #8B,#4D,#10,         -- mov    ecx,count        ; loop counter
   #8B,#55,#14,         -- mov    edx,fcase        ; case sen flag
                        -- @@:
   #AC,                 -- lodsb
   #8A,#1F,             -- mov    bl,byte ptr[edi]
   #47,                 -- inc    edi
   #85,#D2,             -- test   edx,edx
   #75,#1C,             -- jnz    lc
                        -- cbyte:
   #38,#D8,             -- cmp    al,bl
   #75,#0A,             -- jne    @F
   #49,                 -- dec    ecx
   #75,#F1,             -- jnz    @B
   #B8,#01,#00,#00,#00, -- mov    eax,1
   #EB,#05,             -- jmp    done
                        -- @@:
   #B8,#00,#00,#00,#00  -- mov    eax,0
                        -- done:
} &
LEAVE_PROC(16) &
{
                        -- lc:
   #3C,#61,             -- cmp    al, 'a'
   #72,#06,             -- jb     @F
   #3C,#7A,             -- cmp    al, 'z'
   #77,#02,             -- ja     @F
   #24,#DF,             -- and    al,0dfh          ; upper case
                        -- @@:
   #80,#FB,#61,         -- cmp    bl, 'a'
   #72,#08,             -- jb     @F
   #80,#FB,#7A,         -- cmp    bl, 'z'
   #77,#03,             -- ja     @F
   #80,#E3,#DF,         -- and    bl,0dfh
                        -- @@:
   #EB,#CB              -- jmp    cbyte            ; jmp is faster than call/ret
}
global constant CMPS = define_asm_func(code_string,{DWORD,DWORD,DWORD,DWORD},DWORD,4)

global function cmps(atom addr1, atom addr2, atom ccnt,integer f)
 -- addr1 addr2 strings to compare 
 -- ccnt number of bytes to compare
 -- f 0 = case sensitive
    return c_func(CMPS,{addr1,addr2,ccnt,f})
end function

--------------------------------------------------------------------------

global function finds(sequence s, atom addr, atom ccnt, integer case)
 atom a,lmt,cs
 integer c
    c = s[1]  cs = length(s)
    a = allocate(cs)   poke(a,s)
    lmt = addr+ccnt-cs
    while addr <= lmt do
        addr = c_func(CHAR_SCAN,{c,addr,lmt-addr,case})
        if not addr then return 0 end if
        if c_func(CMPS,{a,addr,cs,case}) then return addr end if
        addr += 1
    end while
    return 0
end function

--------------------------------------------------------------------------
-- hash code generator for strings
--------------------------------------------------------------------------

-- not a great hash but won't generate the same code for "elsif" "elsof"

code_string =       -- addr:DWORD, ccnt:DWORD
{
   #55,                 -- push   ebp          ; create stack frame
   #8B,#EC,             -- mov    ebp,esp
   #51,                 -- push   ecx
   #53,                 -- push   ebx
   #56,                 -- push   esi


   #8B,#75,#08,         -- mov    esi,[ebp+8]  ; addr string
   #8B,#4D,#0C,         -- mov    ecx,[ebp+12] ; ccnt
   #33,#C0,             -- xor    eax,eax      ; clear reg
   #8B,#D8,             -- mov    ebx,eax
                        -- @@:
   #AC,                 -- lodsb
   #C1,#C3,#04,         -- rol    ebx,4
   #33,#D8,             -- xor    ebx,eax
   #49,                 -- dec    ecx
   #75,#F7,             -- jnz    @B
   #8B,#C3,             -- mov    eax,ebx


   #5E,                 -- pop    esi
   #5B,                 -- pop    ebx
   #59,                 -- pop    ecx
   #C9,                 -- leave                ; destroy stack frame
-- #25,#FF,#FF,#FF,#3F, -- and    eax,3FFFFFFFh ; return bits 0-29 
   #C2,#08              -- ret    8
}

-- If you need an integer return you can uncomment the next to last line
-- or use remainder(key,#40000000) this will limit the return to positive
-- euphoria integers. 
-- The chances of getting a bad (duplicate) code will increase a little.

-- notes: Euphoria find() is very fast for searching sequences of integers or 
--        strings and fairly slow for searching a sequence of atoms.
--        Unless you do a lot of searches or very large searches the difference
--        in time is probably not significant.
--        A good compromise would be to limit key values to euphoria integers
--        and store the keys in an integer sequence. Using dword_scan() to search
--        values poked into memory will not yield any real preformance improvement
--        over find(), the only advantage would be the ability to use full 32 bit
--        hash codes.

-- Euphoria 2.5, AMD XP2500+
-- Time to generate 1 000 000 hash codes for a 14 byte string .4 sec
-- Time to generate 1 000 000 hash codes for a 256 byte string 1.8 sec
-- Time is for 1000 loops
-- dword_scan() 4.9 sec
--    search a 4 000 000 byte block of memory for a value not in memory
-- find() integer 5.6 sec
--    search 1 000 000 cell sequence for a value not in the sequence
-- find() atom 31.7 sec
--    search 1 000 000 cell sequence for a value not in the sequence
-- find() sequence worst case 85.1 sec
--    search 1 000 000 cell sequence 26 chars per cell all cells identical,
--    find string 26 chars different by last char
-- find() sequence 24.2 sec
--    search 1 000 000 cell sequence 26 chars per cell all cells identical,
--    find string 26 chars different by first char
-- find() sequence best case 6.0 sec
--    search 1 000 000 cell sequence 26 chars per cell all cells identical,
--    find string 1 char
-- find() sequence typical 13.3 sec
--    random strings read from disk 1 002 602 cell sequence 33 353 460 bytes,
--    find string 1 char not in the sequence
-- find() sequence typical 13.2 sec
--    random strings read from disk 1 002 602 cell sequence 33 353 460 bytes,
--    find string 14 chars not in the sequence

global constant HASH_STRING = define_asm_func(code_string,{DWORD,DWORD},DWORD,4)

global function hash_string(atom addr, integer ccnt)
    return c_func(HASH_STRING,{addr,ccnt})
end function

--constant sbuf = allocate(512)
--global function hash(sequence s)
--     poke(sbuf,s)
--     return c_func(HASH_STRING,{sbuf,length(s)})
--end function



